from unittest.mock import call, patch

from django.core.management import call_command
from django.db.migrations import RunPython, RunSQL
from django.test import TestCase

from .commands.collect_sql import SqlObjectOperation, SqlType


class MockMigration(object):
    def __init__(self, operations):
        self.operations = operations


def mock_run_python(apps, schema_editor):
    pass


class CollectSqlTest(TestCase):
    @patch("smartmin.management.commands.collect_sql.Command.load_migrations")
    @patch("smartmin.management.commands.collect_sql.Command.write_dump")
    def test_command(self, mock_write_dump, mock_load_migrations):
        mock_load_migrations.return_value = [
            MockMigration(
                operations=[
                    RunSQL(
                        """
CREATE INDEX test_1 ON foo(bar);
CREATE INDEX test_2 ON foo(bar); create unique index test_3 on foo(bar);
"""
                    ),
                    RunPython(mock_run_python),
                ]
            ),
            MockMigration(
                operations=[
                    RunSQL("DROP INDEX test_2;"),
                ]
            ),
            MockMigration(
                operations=[
                    RunSQL("CREATE TRIGGER test_1 AFTER TRUNCATE ON flows_flowstep EXECUTE PROCEDURE foo();"),
                    RunSQL("CREATE INDEX a_test ON foo(bar);"),
                    RunPython(mock_run_python),
                ]
            ),
            MockMigration(
                operations=[
                    RunSQL(
                        "CREATE OR REPLACE FUNCTION a_func(i INTEGER, s TEXT) RETURNS integer AS $$ BEGIN RETURN i + 1; END; $$ LANGUAGE plpgsql;"
                    ),
                    # different function because it has different parameters
                    RunSQL(
                        "CREATE OR REPLACE FUNCTION a_func(s text, integer) RETURNS text AS $$ BEGIN RETURN UPPER(s); END; $$ LANGUAGE plpgsql;"
                    ),
                    # same function because it has same name and parameters
                    RunSQL(
                        "CREATE OR REPLACE FUNCTION A_FUNC(N integer, V text) RETURNS text AS $$ BEGIN RETURN UPPER(V); END; $$ LANGUAGE plpgsql;"
                    ),
                    RunSQL(
                        "CREATE OR REPLACE FUNCTION func2(i INTEGER) RETURNS integer AS $$ BEGIN RETURN i + 1; END; $$ LANGUAGE plpgsql;"
                    ),
                    RunSQL("DROP FUNCTION func2(i INTEGER);"),
                ]
            ),
        ]

        call_command("collect_sql", output_dir="sql")

        mock_write_dump.assert_has_calls(
            [
                call(
                    "indexes",
                    [
                        SqlObjectOperation(
                            "CREATE INDEX a_test ON foo(bar);",
                            sql_type=SqlType.INDEX,
                            obj_name="a_test",
                            is_create=True,
                        ),
                        SqlObjectOperation(
                            "CREATE INDEX test_1 ON foo(bar);",
                            sql_type=SqlType.INDEX,
                            obj_name="test_1",
                            is_create=True,
                        ),
                        SqlObjectOperation(
                            "create unique index test_3 on foo(bar);",
                            sql_type=SqlType.INDEX,
                            obj_name="test_3",
                            is_create=True,
                        ),
                    ],
                    "sql",
                ),
                call(
                    "functions",
                    [
                        SqlObjectOperation(
                            "CREATE OR REPLACE FUNCTION A_FUNC(N integer, V text) RETURNS text AS $$ BEGIN RETURN UPPER(V); END; $$ LANGUAGE plpgsql;",
                            sql_type=SqlType.FUNCTION,
                            obj_name="A_FUNC(integer,text)",
                            is_create=True,
                        ),
                        SqlObjectOperation(
                            "CREATE OR REPLACE FUNCTION a_func(s text, integer) RETURNS text AS $$ BEGIN RETURN UPPER(s); END; $$ LANGUAGE plpgsql;",
                            sql_type=SqlType.FUNCTION,
                            obj_name="a_func(text,integer)",
                            is_create=True,
                        ),
                    ],
                    "sql",
                ),
                call(
                    "triggers",
                    [
                        SqlObjectOperation(
                            "CREATE TRIGGER test_1 AFTER TRUNCATE ON flows_flowstep EXECUTE PROCEDURE foo();",
                            sql_type=SqlType.TRIGGER,
                            obj_name="test_1",
                            is_create=True,
                        ),
                    ],
                    "sql",
                ),
            ]
        )



from django.db import migrations, models

import smartmin.csv_imports.models


class Migration(migrations.Migration):
    dependencies = [
        ("csv_imports", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="importtask",
            name="csv_file",
            field=models.FileField(
                help_text="A comma delimited file of records to import",
                upload_to=smartmin.csv_imports.models.generate_file_path,
                verbose_name="Import file",
            ),
        ),
    ]



import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("csv_imports", "0003_importtask_task_status"),
    ]

    operations = [
        migrations.AlterField(
            model_name="importtask",
            name="created_on",
            field=models.DateTimeField(
                blank=True,
                default=django.utils.timezone.now,
                editable=False,
                help_text="When this item was originally created",
            ),
        ),
        migrations.AlterField(
            model_name="importtask",
            name="modified_on",
            field=models.DateTimeField(
                blank=True,
                default=django.utils.timezone.now,
                editable=False,
                help_text="When this item was last modified",
            ),
        ),
    ]



import django.views.static
from django.conf import settings
from django.http import HttpResponseRedirect
from django.urls import reverse

from .models import PasswordHistory


class ChangePasswordMiddleware:
    """
    Redirects all users to the password change form if we find that a user's
    password is expired.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

        self.password_expire = getattr(settings, "USER_PASSWORD_EXPIRATION", -1)

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, view, *args, **kwargs):
        newpassword_path = reverse("users.user_newpassword", args=[0])
        logout_path = reverse("users.user_logout")

        if (
            self.password_expire < 0
            or not request.user.is_authenticated
            or view == django.views.static.serve
            or request.path == newpassword_path
            or request.path == logout_path
        ):  # noqa
            return

        if PasswordHistory.is_password_expired(request.user):
            return HttpResponseRedirect(reverse("users.user_newpassword", args=["0"]))



import re

from django.conf import settings
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.management import create_contenttypes
from django.contrib.contenttypes.models import ContentType

perm_desc_regex = re.compile(r"(?P<app>\w+)\.(?P<contenttype>[a-z0-9]+)((_(?P<perm>\w+))|(\.(?P<wild>\*)))")


def update_group_permissions(app, group, permissions: list):
    """
    Checks the the passed in role (can be user, group or AnonymousUser)  has all the passed
    in permissions, granting them if necessary.
    """

    new_permissions = []

    for perm_desc in permissions:
        app_label, content_type, perm = _parse_perm_desc(perm_desc)

        # ignore if this permission is for a type not in this app
        if app_label != app.label:
            continue

        if perm == "*":
            codenames = Permission.objects.filter(
                content_type__app_label=app_label, codename__startswith=f"{content_type}_"
            ).values_list("codename", flat=True)
        else:
            codenames = [f"{content_type}_{perm}"]

        perms = []
        for codename in codenames:
            try:
                perms.append(Permission.objects.get(content_type__app_label=app_label, codename=codename))
            except Permission.DoesNotExist:
                raise ValueError(f"Cannot grant permission {app_label}.{codename} as it does not exist.")

            new_permissions.append((app_label, codename))

        group.permissions.add(*perms)

    # remove any that are extra
    for perm in group.permissions.filter(content_type__app_label=app.label):
        if (perm.content_type.app_label, perm.codename) not in new_permissions:
            group.permissions.remove(perm)


def sync_permissions(sender, **kwargs):
    """
    1. Ensures all permissions for this app described by the PERMISSIONS setting exist in the database.
    2. Ensures all permissions for this app granted by the GROUP_PERMISSIONS setting are granted.
    """

    # the content types app also listens for post_migrate signals but since order isn't guaranteed, we need to
    # manually invoke what it does for this app to be sure that the content types are created
    create_contenttypes(sender)

    # for each of our items
    for natural_key, permissions in getattr(settings, "PERMISSIONS", {}).items():
        # if wild, we add these permissions to all content types defined by this app
        if natural_key == "*":
            for content_type in ContentType.objects.filter(app_label=sender.label):
                for permission in permissions:
                    _ensure_permission_exists(content_type, permission)

        # otherwise check if this type belongs to this app and if so add the permissions to that type only
        else:
            app_label, model = natural_key.split(".")
            if app_label == sender.label:
                try:
                    content_type = ContentType.objects.get_by_natural_key(app_label, model)
                except ContentType.DoesNotExist:
                    raise ValueError(f"No such content type: {app_label}.{model}")

                # add each permission
                for permission in permissions:
                    _ensure_permission_exists(content_type, permission)

    # for each of our items
    for name, permissions in getattr(settings, "GROUP_PERMISSIONS", {}).items():
        # get or create the group
        group, created = Group.objects.get_or_create(name=name)

        update_group_permissions(sender, group, permissions)


def _parse_perm_desc(desc: str) -> tuple:
    """
    Parses a permission descriptor into its app_label, model and permission parts, e.g.
        app.model.*      => app, model, *
        app.model_perm   => app, model, perm
    """

    match = perm_desc_regex.match(desc)
    if not match:
        raise ValueError(f"Invalid permission descriptor: {desc}")

    return match.group("app"), match.group("contenttype"), match.group("perm") or match.group("wild")


def _ensure_permission_exists(content_type, permission: str):
    """
    Adds the passed in permission to that content type.  Note that the permission passed
    in should be a single word, or verb.  The proper 'codename' will be generated from that.
    """

    codename = f"{content_type.model}_{permission}"  # build our permission slug

    Permission.objects.get_or_create(
        content_type=content_type, codename=codename, defaults={"name": f"Can {permission} {content_type.name}"}
    )



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="ImportTask",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                (
                    "is_active",
                    models.BooleanField(
                        default=True, help_text="Whether this item is active, use this instead of deleting"
                    ),
                ),
                (
                    "created_on",
                    models.DateTimeField(help_text="When this item was originally created", auto_now_add=True),
                ),
                ("modified_on", models.DateTimeField(help_text="When this item was last modified", auto_now=True)),
                (
                    "csv_file",
                    models.FileField(
                        help_text="A comma delimited file of records to import",
                        upload_to="csv_imports",
                        verbose_name="Import file",
                    ),
                ),
                ("model_class", models.CharField(help_text="The model we are importing for", max_length=255)),
                (
                    "import_params",
                    models.TextField(help_text="JSON blob of form parameters on task creation", null=True, blank=True),
                ),
                ("import_log", models.TextField()),
                (
                    "import_results",
                    models.TextField(help_text="JSON blob of result values on task completion", null=True, blank=True),
                ),
                ("task_id", models.CharField(max_length=64, null=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        related_name="csv_imports_importtask_creations",
                        on_delete=models.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        help_text="The user which originally created this item",
                    ),
                ),
                (
                    "modified_by",
                    models.ForeignKey(
                        related_name="csv_imports_importtask_modifications",
                        on_delete=models.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                        help_text="The user which last modified this item",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.forms import widgets
from django.utils.html import escape
from django.utils.safestring import mark_safe


class VisibleHiddenWidget(widgets.Widget):
    def render(self, name, value, attrs=None, renderer=None):
        """
        Returns this Widget rendered as HTML, as a Unicode string.

        The 'value' given is not guaranteed to be valid input, so subclass
        implementations should program defensively.
        """
        html = ""
        html += "%s" % value
        html += '<input type="hidden" name="%s" value="%s">' % (escape(name), escape(value))
        return mark_safe(html)


class DatePickerWidget(widgets.DateInput):
    """
    Date input which uses Javascript date picker widget
    """

    input_format = ("MM d, yyyy", "%B %d, %Y")  # Javascript and Python format strings

    def __init__(self, *args, **kwargs):
        kwargs["attrs"] = {"data-provide": "datepicker", "data-date-format": self.input_format[0]}
        kwargs["format"] = self.input_format[1]

        super(DatePickerWidget, self).__init__(*args, **kwargs)

    class Media:
        js = ("js/bootstrap-datepicker.js",)
        css = {"all": ("css/bootstrap-datepicker3.css",)}


class ImageThumbnailWidget(widgets.ClearableFileInput):
    def __init__(self, thumb_width=75, thumb_height=75):
        self.width = thumb_width
        self.height = thumb_height
        super(ImageThumbnailWidget, self).__init__({})

    def render(self, name, value, attrs=None, renderer=None):
        thumb_html = "<table><tr>"
        if value and hasattr(value, "url"):
            try:
                from sorl.thumbnail import get_thumbnail

                value = get_thumbnail(value, f"{self.width}x{self.height}", crop="center", quality=99)
            except ImportError:
                pass

            thumb_html += '<td><img src="%s" width="%s" width="%s" /></td>' % (value.url, self.width, self.height)

        thumb_html += '<td><input type="checkbox" name="%s-clear" /> Clear' % name
        thumb_html += '<input type="file" name="%s" /></td>' % name
        thumb_html += "</tr></table>"

        return mark_safe(str('<div class="image-picker">%s</div>' % thumb_html))



from django.utils.module_loading import import_string

from smartmin.csv_imports.models import ImportTask
from smartmin.views import SmartCRUDL, SmartListView, SmartReadView


class ImportTaskCRUDL(SmartCRUDL):
    model = ImportTask
    actions = ("read", "list")

    class Read(SmartReadView):
        def derive_refresh(self):
            if self.object.status() in ["PENDING", "RUNNING", "STARTED"]:
                return 2000
            else:
                return 0

    class List(SmartListView):
        fields = ("status", "type", "csv_file", "created_on", "created_by")
        link_fields = ("csv_file",)

        def get_type(self, obj):
            return import_string(obj.model_class)._meta.verbose_name_plural.title()
